@@ -6,8 +6,9 @@ from collections import OrderedDict
 import re
 
 from dateutil.parser import parse
+from datetime import timedelta
 
-from blaseball_mike import database, reference, tables
+from blaseball_mike import database, reference, chronicler, tables
 
 
 
@@ -23,11 +24,15 @@ class Base(abc.ABC):
 
     @staticmethod
     def _camel_to_snake(name):
-        return Base._camel_to_snake_re.sub('_', name).lower()
+        return Base._remove_leading_underscores(Base._camel_to_snake_re.sub('_', name)).lower()
+
+    @staticmethod
+    def _remove_leading_underscores(name):
+        return name.strip('_')
 
     def json(self):
         return {
-            f: getattr(self, self._camel_to_snake(f)) for f in self.fields
+            f: getattr(self, self._remove_leading_underscores(self._camel_to_snake(f))) for f in self.fields
         }
 
 
@@ -101,6 +106,22 @@ class Player(Base):
         """
         return cls.load(id_).get(id_)
 
+    @classmethod
+    def load_one_at_time(cls, id_, time):
+        if isinstance(time, str):
+            time = parse(time)
+
+        players = chronicler.get_player_updates(id_, before=time, order="desc", count=1)
+        return cls(dict(players.get(id_)["data"], **{"timestamp":time}))
+
+    @classmethod
+    def load_history(cls, id_, order='desc'):
+        """
+        Returns array of Player stat changes with most recent first.
+        """
+        players = chronicler.get_player_history(id_, order=order)
+        return [cls(dict(p['data'], timestamp=p['firstSeen'])) for p in players]
+
     @classmethod
     def find_by_name(cls, name):
         """
@@ -428,12 +449,23 @@ class Team(Base):
                 return team
         return None
 
+    @classmethod
+    def load_at_time(cls, id_, time):
+        if isinstance(time, str):
+            time = parse(time)
+
+        team = chronicler.get_team_updates(id_, before=time, order="desc", count=1)
+        return cls(dict(team.get(id_)["data"], **{"timestamp": time}))
+
     @property
     def lineup(self):
         if self._lineup:
             return self._lineup
-        players = Player.load(*self._lineup_ids)
-        self._lineup = [players.get(id_) for id_ in self._lineup_ids]
+        if getattr(self, "timestamp", None):
+            self._lineup = [Player.load_one_at_time(x, self.timestamp) for x in self._lineup_ids]
+        else:
+            players = Player.load(*self._lineup_ids)
+            self._lineup = [players.get(id_) for id_ in self._lineup_ids]
         return self._lineup
 
     @lineup.setter
@@ -445,8 +477,11 @@ class Team(Base):
     def rotation(self):
         if self._rotation:
             return self._rotation
-        players = Player.load(*self._rotation_ids)
-        self._rotation = [players.get(id_) for id_ in self._rotation_ids]
+        if getattr(self, "timestamp", None):
+            self._rotation = [Player.load_one_at_time(x, self.timestamp) for x in self._rotation_ids]
+        else:
+            players = Player.load(*self._rotation_ids)
+            self._rotation = [players.get(id_) for id_ in self._rotation_ids]
         return self._rotation
 
     @rotation.setter
@@ -458,8 +493,11 @@ class Team(Base):
     def bullpen(self):
         if self._bullpen:
             return self._bullpen
-        players = Player.load(*self._bullpen_ids)
-        self._bullpen = [players.get(id_) for id_ in self._bullpen_ids]
+        if getattr(self, "timestamp", None):
+            self._bullpen = [Player.load_one_at_time(x, self.timestamp) for x in self._bullpen_ids]
+        else:
+            players = Player.load(*self._bullpen_ids)
+            self._bullpen = [players.get(id_) for id_ in self._bullpen_ids]
         return self._bullpen
 
     @bullpen.setter
@@ -471,8 +509,11 @@ class Team(Base):
     def bench(self):
         if self._bench:
             return self._bench
-        players = Player.load(*self._bench_ids)
-        self._bench = [players.get(id_) for id_ in self._bench_ids]
+        if getattr(self, "timestamp", None):
+            self._bench = [Player.load_one_at_time(x, self.timestamp) for x in self._bench_ids]
+        else:
+            players = Player.load(*self._bench_ids)
+            self._bench = [players.get(id_) for id_ in self._bench_ids]
         return self._bench
 
     @bench.setter
@@ -512,6 +553,14 @@ class Team(Base):
     def game_attr(self, value):
         self._game_attr = value
 
+    @property
+    def card(self):
+        return tables.Tarot(self._card)
+
+    @card.setter
+    def card(self, value):
+        self._card = value
+
 
 class Division(Base):
 
@@ -534,7 +583,7 @@ class Division(Base):
         Name can be full name or nickname, case insensitive.
         """
         divisions = cls.load_all()
-        for division in divisions:
+        for division in divisions.values():
             if name in division.name:
                 return division
         return None
@@ -645,6 +694,12 @@ class Game(Base):
             id_: cls(game) for id_, game in database.get_games(season, day).items()
         }
 
+    @classmethod
+    def load_by_season(cls, season, team_id=None, day=None):
+        return {
+            id_: cls(game["data"]) for id_, game in chronicler.get_games(team_ids=team_id, season=season, day=day).items()
+        }
+
     @property
     def winning_team(self):
         return self.home_team if self.home_score > self.away_score else self.away_team
@@ -961,6 +1016,22 @@ class BlessingResult(Base):
         return self.bonus_id
 
 
+class TidingResult(Base):
+
+    @classmethod
+    def load(cls, *ids):
+        event = database.get_offseason_event_results(list(ids))
+        return {
+            id_: cls(event) for (id_, event) in event.items()
+        }
+
+    @classmethod
+    def load_one(cls, id_):
+        return cls.load(id_).get(id_)
+
+EventResult = TidingResult
+
+
 class ElectionResult(Base):
 
     @classmethod
@@ -972,7 +1043,7 @@ class ElectionResult(Base):
         if self._bonus_results:
             return self._bonus_results
         if not self._bonus_results_ids:
-            return None
+            return []
         blessings = BlessingResult.load(*self._bonus_results_ids)
         self._bonus_results = [blessings.get(id_) for id_ in self._bonus_results_ids]
         return self._bonus_results
@@ -992,7 +1063,7 @@ class ElectionResult(Base):
         if self._decree_results:
             return self._decree_results
         if not self._decree_results_ids:
-            return None
+            return []
         decrees = DecreeResult.load(*self._decree_results_ids)
         self._decree_results = [decrees.get(id_) for id_ in self._decree_results_ids]
         return self._decree_results
@@ -1002,6 +1073,26 @@ class ElectionResult(Base):
         self._decree_results = None
         self._decree_results_ids = value
 
+    @property
+    def event_results(self):
+        if self._event_results:
+            return self._event_results
+        if not self._event_results_ids:
+            return []
+        events = TidingResult.load(*self._event_results_ids)
+        self._event_results = [events.get(id_) for id_ in self._event_results_ids]
+        return self._event_results
+
+    @event_results.setter
+    def event_results(self, value):
+        self._event_results = None
+        self._event_results_ids = value
+
+    # tiding_results is an alias to event_results
+    @property
+    def tiding_results(self):
+        return self.event_results
+
 OffseasonResult = ElectionResult
 
 
@@ -1271,11 +1362,25 @@ class Tribute(Base):
             tributes_dict[tribute['playerId']] = cls(tribute)
         return tributes_dict
 
+    @classmethod
+    def load_at_time(cls, id_, time):
+        if isinstance(time, str):
+            time = parse(time)
+
+        tributes = chronicler.get_tribute_updates(before=time, order="desc", count=1)
+        tributes_dict = OrderedDict()
+        for tribute in tributes:
+            tributes_dict[tribute['player_id']] = cls(dict(tribute.get(id_)["data"], **{"timestamp": time}))
+        return tributes_dict
+
     @property
     def player(self):
         if getattr(self, '_player', None):
             return self._player
-        self._player = Player.load_one(self.player_id)
+        if getattr(self, "timestamp", None):
+            self._player = Player.load_one_at_time(self.player_id, self.timestamp)
+        else:
+            self._player = Player.load_one(self.player_id)
         return self._player
 
     @player.setter
@@ -1327,6 +1432,7 @@ class SeasonStatsheet(Base):
             stats_dict[k] = cls(v)
         return stats_dict
 
+    @classmethod
     def load_by_season(cls, season):
         """Season is 1 indexed."""
         season = Season.load(season)
